(*************************************)
(**                                 **)
(**   ZVID VIEW   by Zatzen 2014    **)
(**                                 **)
(**   this version supports         **)
(**   file format v1.0              **)
(**                                 **)
(*************************************)

{$G+}

unit zvidview;

interface

function load_zvid(n: string; slot: byte): byte;
procedure init_zvid(n, xdim, ydim: word; dest: pointer);
procedure zvid_putframe(slot: byte; frame, x, y: word);
function zvid_frames(slot: byte): word;

implementation

const

  version_main = 1;
  version_sub  = 0;


type Tzvidheader = record
  id: longint; { $4449565A }
  chunksize: longint;
  ver_main: byte;
  ver_sub: byte;
end;

type Tzvidheaderinfo = record
  frames: word;
  redblocksize: array[0..3] of word;
  palsize: word;
end;

type Tframeslist = record
  xdim: byte;
  ydim_transp: byte;
  transp_color: byte;
end;

type frameslist_array = array[0..8191] of Tframeslist;

type word_array = array[0..32766] of word;

type byte_array = array[0..65534] of byte;

type pointer_array = array[0..16382] of pointer;

type zvidinfo_array = array[0..255] of Tzvidheaderinfo;

type pointer_direct = array[0..1] of word;

var

  p_out: pointer;
  out_buffer: ^byte_array absolute p_out;

  outblkbuf_pointer: pointer;
  outblkbuf: ^byte_array absolute outblkbuf_pointer;

  {
  pal: array[0..31] of byte;
  pal_pointer: pointer;
  }

  bitdata_pointer: pointer;
  bitdata: ^byte_array absolute bitdata_pointer;

{  bitdata: array[0..71] of byte;} {40 + 32 inklusive pal an offset 40}


  {bitpos: word;}


  zvidheader: Tzvidheader; { wird jeweils nur 1x beim laden gebraucht }

  zvidinfo_pointer: pointer;
  zvidinfo: ^zvidinfo_array absolute zvidinfo_pointer;
  { direkt adressierbar fr alle slots }

  frameslist_pointer_array: pointer;
  frameslist_pointer: ^pointer_array absolute frameslist_pointer_array;
  { zeiger array mit pointern die jeweils }
  { auf die frameslist fr jeden slot zeigen (max. 256) }

  frameslist_blockdatasize_pointer_array: pointer;
  frameslist_blockdatasize_pointer:
    ^pointer_array absolute frameslist_blockdatasize_pointer_array;

  p1, p2: pointer;

  frameslist_blockdatasize: ^word_array absolute p1; {direkter zugriff}
  frameslist: ^frameslist_array absolute p2;         {direkter zugriff}


  blockdata_pointer_pointer_array: pointer;
  blockdata_pointer_pointer:
    ^pointer_array absolute blockdata_pointer_pointer_array;

  p3: pointer;

  blockdata_pointer: ^pointer_array absolute p3;
  { direkter zugriff auf die pointer zu den blockdaten }


  p4: pointer;

  blockdata: ^byte_array absolute p4;

   { direkter zugriff }



  redblockdata_pointer_array: pointer;
  redblockdata_pointer: ^pointer_array absolute redblockdata_pointer_array;
  {max. 1024 baenke}

  p5: pointer;

  redblockdata: ^byte_array absolute p5;


  paldata_pointer_array: pointer;
  paldata_pointer: ^pointer_array absolute paldata_pointer_array;

  {max. 256 verschiedene paletten}

  p6: pointer;
  paldata: ^byte_array absolute p6;


  block_transparent: bytebool;


  transp_col: byte;


function zvid_frames(slot: byte): word;
begin
  zvid_frames := zvidinfo^[slot].frames;
end;

procedure init_zvid(n, xdim, ydim: word; dest: pointer);
 { Pointer Datenfelder anlegen, n max. 256 }
  var
    i: byte;
begin
  { n ist die anzahl einzurichtender slots (anzahl der zvid-dateien) }

{  bitdata_pointer := addr(bitdata^[0]);}

  getmem(outblkbuf_pointer, 80);{16 Byte mehr zum Ausrichten auf Null Offset}


  outblkbuf_pointer := ptr(seg(outblkbuf_pointer^)
    + ofs(outblkbuf_pointer^) div 16 + 1, 0);


  getmem(bitdata_pointer, 72);

  {writeln(ofs(outblkbuf_pointer^));}

  p_out := dest;

  getmem(zvidinfo_pointer, n * sizeof(Tzvidheaderinfo));

  { pointer-array fr n frameslists reservieren }
  getmem(frameslist_pointer_array, n * 4);

  { pointer-array fr n blockdatengre-listen reservieren }
  getmem(frameslist_blockdatasize_pointer_array, n * 4);

  { pointer-array fr n pointer arrays fr blockdaten reservieren }
  getmem(blockdata_pointer_pointer_array, n * 4);

  getmem(redblockdata_pointer_array, n * 4 * 4); { 4 Baenke }

  getmem(paldata_pointer_array, n * 4);

end;



function load_zvid(n: string; slot: byte): byte;
  var
    bank: byte;
    frame: word;
    f: file;
    inp: string;
    ch: char;
begin

  load_zvid := 0;

  assign(f, n);
  {$I-}
  reset(f, 1);
  {$I+}
  if ioresult <> 0 then
  begin
    {writeln('error opening ', n);}
    load_zvid := 1;
    exit;
  end;

  blockread(f, zvidheader, sizeof(zvidheader));

  if zvidheader.id <> $4449565A then
  begin
    close(f);
    {writeln(n, ' is not a valid ZVID file (wrong ID)');}
    load_zvid := 2;
    exit;
  end;

  with zvidheader do
  if (ver_main <> version_main) or (ver_sub <> version_sub) then
  begin
    {
    writeln('ZVID file ', n, ' is version ', ver_main, '.', ver_sub,
      ' and not version ', version_main, '.', version_sub);
    write(chr(7), 'continue anyway? (Y/N): ');
    ch := '_';
    while (ch <> 'Y') and (ch <> 'N') do
    begin
      readln(inp);
      ch := upcase(inp[1]);
    end;
    if ch = 'N' then
    begin
      close(f);
      writeln('program aborted');
      halt(2);
    end;
    }
    close(f);
    load_zvid := 3;
    exit;
  end;

  blockread(f, zvidinfo^[slot], sizeof(Tzvidheaderinfo));



  (*** JETZT DIE FRAMESLIST LADEN ***)


  (*** Erst Blockdatasizes ***)

  getmem(frameslist_blockdatasize_pointer^[slot],
    zvidinfo^[slot].frames * 2);
  { einen pointer, fr diesen slot, reservieren }

  p1:= frameslist_blockdatasize_pointer^[slot];

  {p1 zeigt nun auf den anfang der blockdatasize-liste fuer den jeweil. slot}


{  writeln('Framelist-Blocklength-Data -- frames * 2: ', zvidinfo^[slot].frames * 2);}
  blockread(f, p1^, zvidinfo^[slot].frames * 2);

  { Blockdatasizes fr alles Frames in diesem Slot geladen wo p1 hinzeigt }

  (*** Nach komplettem Laden des ZVD mit freemem wieder freigeben ! ***)


  (*** Jetzt frameslist ***)

  getmem(frameslist_pointer^[slot],
    zvidinfo^[slot].frames * sizeof(Tframeslist));

  { Element frameslist_pointer^[slot]^ zeigt nun auf die aktuelle Frameslist }

  p2 := frameslist_pointer^[slot];

{  writeln('Frameslist x-y-transpcol: ', zvidinfo^[slot].frames * sizeof(Tframeslist));}
  blockread(f, p2^, zvidinfo^[slot].frames * sizeof(Tframeslist));

  { p2 zeigt auf das Frameslist-Datenfeld dieses Slots }

  { Zugriff ber frameslist^[].xxx }

  { p1 zeigt auf das Datenfeld der Frame-Blocksizes (WORDs) }

  { Zugriff ber frames_blockdatasize^[] }


  { Frame-Daten Speicher reservieren und Blcke laden }


  { p1 und p2 zeigen auf Info-Daten des aktuellen Slots }


  { Pointer-Datenfeld anlegen fr frames * blockdaten }


  getmem(blockdata_pointer_pointer^[slot], 4 * zvidinfo^[slot].frames);
  { pointerdatenfeld fr diesen slot angelegt, ein pointer fr jeden frame }

  p3 := blockdata_pointer_pointer^[slot];


  { Daten, die hier nur geladen aber nicht ausgelesen }
  { werden mssen bentigen keinen px pointer }

  for frame := 0 to zvidinfo^[slot].frames - 1 do
  begin

    getmem(blockdata_pointer^[frame], frameslist_blockdatasize^[frame]);

    { p4 := blockdata_pointer^[frame]; }
    {blockread(f, p4^, frameslist_blockdatasize^[frame]);}
{    writeln('frame #', frame, ': ', frameslist_blockdatasize^[frame]);}
    blockread(f, blockdata_pointer^[frame]^,
      frameslist_blockdatasize^[frame]);

  end;


  (*** WEITER MIT DEN REDUNDANTEN BLOCKDATEN ***)

  for bank := 0 to 3 do
    if zvidinfo^[slot].redblocksize[bank] > 0 then
    begin

      getmem(redblockdata_pointer^[slot * 4 + bank],
        zvidinfo^[slot].redblocksize[bank]);

{      writeln('redblock bank #', bank, ': ', zvidinfo^[slot].redblocksize[bank]);}
      blockread(f, redblockdata_pointer^[slot * 4 + bank]^,
        zvidinfo^[slot].redblocksize[bank]);

    end;


  (*** WEITER MIT Palette ***)

  getmem(paldata_pointer^[slot], zvidinfo^[slot].palsize);

{  writeln('Palette Data: ', zvidinfo^[slot].palsize);}
  blockread(f, paldata_pointer^[slot]^, zvidinfo^[slot].palsize);



  (*** alles geladen ***)



  (*** Speicher von Daten freigeben die nur zum Laden bentigt wurden ***)

  freemem(frameslist_blockdatasize_pointer^[slot],
    zvidinfo^[slot].frames * 2);


  close(f);

end;


procedure decode_bitstream_asm_1; assembler;
asm

  push ds

  les di, outblkbuf_pointer
  lds si, bitdata_pointer

  mov dx, ds:[si+40] { 2 Farben Palette in DX laden }

  xor ah, ah
  mov bx, 8

  @outerloop:
    lodsb
    mov cx, bx
    inc ah

    @loop:

      shr al, 1
      jnc @null
      mov es:[di], dh
      jmp @done
      @null:
      mov es:[di], dl
      @done:
      inc di

    dec cx
    jnz @loop

  test ah, bl
  jz @outerloop

  pop ds

end;




procedure decode_bitstream_asm_2; assembler;
asm

  push ds

  les di, outblkbuf_pointer
  lds si, bitdata_pointer

  db 66h; mov bx, ds:[si+40] { Farben 0-3 in EBX laden }

  xor cx, cx

  @outerloop:

    inc ch
    xor ah, ah
    lodsb

    @loop:

      inc ah

      db 66h; mov dx, bx { EDX mit EBX beladen }
      mov cl, al
      and cl, 3          { cl enthlt nun index auf palette: 0, 1, 2 oder 3 }
      shl cl, 3          { mit 8 multiplizieren }
      db 66h; shr dx, cl { EDX um cl bits rechtsshiften -> Farbnummer in DL }
      mov es:[di], dl
      inc di
      shr al, 2          { al eine Position weiter nach rechts }

    test ah, 4
    jz @loop

  test ch, 16
  jz @outerloop


  pop ds

end;



procedure decode_bitstream_asm_3; assembler;
asm

  push ds

  les di, outblkbuf_pointer
  lds si, bitdata_pointer

  db 66h; mov bx, ds:[si+40] { Farben 0-3 in EBX }
  db 66h; mov dx, ds:[si+44] { Farben 4-7 in EDX }

  xor ch, ch
  @outerloop:

    mov ax, ds:[si+1] { untere 3 Byte von EAX... }
    db 66h; shl ax, 8 { ... }
    mov al, ds:[si]   { ... mit Daten fr 8 Pixel (= 1 Zeile) beladen }
    add si, 3

  @loop:

    mov cl, al
    shl cl, 6  { oberste 6 Bit rausschieben, Bit 2 landen im CF }

    jc @edx  { Carry gesetzt -> Farbe 4-7 -> EBX }
      shr cl, 3 { 3 Bit nach rechts, dadurch die brigen 2 Bit mit 8 mult. }
      db 66h; ror bx, cl { EBX um CL bits rechtsrotieren -> Farbe in BL }
      mov es:[di], bl
      db 66h; rol bx, cl { zurueckrotieren }
      jmp @done
    @edx:
      shr cl, 3
      db 66h; ror dx, cl { s.o. -> Farbe in DL }
      mov es:[di], dl
      db 66h; rol dx, cl { zurueckrotieren }
    @done:
      inc di
      db 66h; shr ax, 3 { EAX 3 Bit weiter schieben }

  inc ch
  test ch, 7
  jnz @loop

  test ch, 64
  jz @outerloop

  pop ds

end;


procedure decode_bitstream_asm_3_table; assembler;
asm
  push ds

  les di, outblkbuf_pointer
  lds si, bitdata_pointer
  lea bx, [si+40] { Anfangsposition der Paletten-Tabelle -> BX }

  mov ah, 7
  mov ch, 8
  @outerloop:

    mov dx, ds:[si+1] { untere 3 Byte von EDX... }
    db 66h; shl dx, 8 { ... }
    mov dl, ds:[si]   { ... mit Daten fr 8 Pixel (= 1 Zeile) beladen }
    add si, 3

    mov cl, 8
    @loop:
      mov al, dl
      and al, ah
      xlat
      stosb
      db 66h; shr dx, 3
    dec cl
    jnz @loop

   dec ch
   jnz @outerloop

  pop ds
end;



procedure decode_bitstream_asm(bits: byte); assembler;
{  bitdata_pointer := addr(bitdata[0]);
   outblkbuf_pointer := addr(outblkbuf[0]);}


asm

  xor ch, ch
  inc ch
  mov cl, bits
  shl ch, cl
  dec ch { maske fertig }


  push ds
  (*********)

  les di, outblkbuf_pointer
  lds si, bitdata_pointer


  xor dx, dx  {bitposition auf null setzen}


  @loop:

  mov bx, dx

  xor ah, ah
  mov al, bits
  add dx, ax { bitpos erhhen fr nchsten durchlauf }

  mov cl, bl { fr bitpos mod 8 <=> bitpos and 7 }

  shr bx, 3 { bytepos in bx }

  mov ax, ds:[si+bx]

(**************)
  and cl, 7   { cx: bitpos -> bitpos mod 8}
  shr ax, cl

  and al, ch
(**************)
  { palettenposition bestimmt }

  mov bl, 40
  add bl, al
  mov al, ds:[si+bx]

  mov es:[di], al
  inc di
(********************)



  test di, 64
  jz @loop


  (*********)

  pop ds

end;



procedure copy_block_asm(x, y: word); assembler;
asm

  push ds

  les di, p_out
  lds si, outblkbuf_pointer

  mov ax, 320
  mov dx, y
  shl dx, 3
  mul dx

  mov dx, x
  shl dx, 3
  add ax, dx
  add di, ax


  mov dx, 320
  sub dx, 8

  mov cx, 8

  @loop:

    movsw
    movsw
    movsw
    movsw

    add di, dx


  dec cx
  jnz @loop


  pop ds

end;


procedure copy_block_transp_asm(x, y: word); assembler;
asm

  mov bl, transp_col

  push ds

  les di, p_out
  lds si, outblkbuf_pointer

  mov ax, 320
  mov dx, y
  shl dx, 3
  mul dx

  mov dx, x
  shl dx, 3
  add ax, dx
  add di, ax


  mov dx, 320
  sub dx, 8

  mov cx, 8

  @loop:

    lodsb
    cmp al, bl
    je @skip0
    mov es:[di], al
    @skip0:
    inc di

    lodsb
    cmp al, bl
    je @skip1
    mov es:[di], al
    @skip1:
    inc di

    lodsb
    cmp al, bl
    je @skip2
    mov es:[di], al
    @skip2:
    inc di

    lodsb
    cmp al, bl
    je @skip3
    mov es:[di], al
    @skip3:
    inc di

    lodsb
    cmp al, bl
    je @skip4
    mov es:[di], al
    @skip4:
    inc di

    lodsb
    cmp al, bl
    je @skip5
    mov es:[di], al
    @skip5:
    inc di

    lodsb
    cmp al, bl
    je @skip6
    mov es:[di], al
    @skip6:
    inc di

    lodsb
    cmp al, bl
    je @skip7
    mov es:[di], al
    @skip7:
    inc di

    add di, dx


  dec cx
  jnz @loop


  pop ds

end;


procedure copy_block(x, y: word);
var
  i: byte;
begin

{  fillchar(outblkbuf[0], 64, 15);}
  for i := 0 to 7 do
    move(outblkbuf^[i*8], out_buffer^[320 * (y+i) + x], 8);

end;


procedure zvid_putframe(slot: byte; frame, x, y: word);
  var
    blocktype, blockheader: byte;
    blkpos, frame_blocks, blkcount: word;
    bitdatasize, bits: byte;
    decpos: word;
    pal_offset, red_offset: word;
    i, blockrepeat: word;

  label
    start, test;
begin

  p2 := frameslist_pointer^[slot];
  frame_blocks := frameslist^[frame].xdim
    * (frameslist^[frame].ydim_transp and 127);

  p3 := blockdata_pointer_pointer^[slot];
  p4 := blockdata_pointer^[frame];
  p6 := paldata_pointer^[slot];

  transp_col := frameslist^[frame].transp_color;
{
  p5 := redblockdata_pointer^[slot * 4 + bank];
}

  blkpos := 0;
  blkcount := 0;

  repeat

    {goto test;}

    blockrepeat := 0;

    start:

    block_transparent := false;

    blockheader := blockdata^[blkpos];
    inc(blkpos);
    blocktype := blockheader and 15;


    if blocktype = 0 then { einfarbig }
    begin

      if blockheader and 64 <> 0 then
      begin
        fillchar(outblkbuf^[0], 64, blockdata^[blkpos]);
        inc(blkpos);
      end;

      if blockheader and 32 <> 0 then
        fillchar(outblkbuf^[0], 64, ((blockheader and 16) shr 4) * 255);

      if blockheader = 0 then block_transparent := true;

    end;


    if (blocktype = 1) or (blocktype = 2) then { bitweise 1, 2, 3, 4, 5 }
    begin

      if blocktype = 2 then block_transparent := true;

      bits := (blockheader shr 4) and 7;
      bitdatasize := bits shl 3;

      move(blockdata^[blkpos], bitdata^[0], bitdatasize);
      inc(blkpos, bitdatasize);

      pal_offset := blockdata^[blkpos];
      inc(blkpos);
      if blockheader and 128 <> 0 then
      begin
        inc(pal_offset, 256 * blockdata^[blkpos]);
        inc(blkpos);
      end;


      move(paldata^[pal_offset], bitdata^[40]{pal[0]}, 1 shl bits);

      {
      bitpos := 0;
      for i := 0 to 63 do
        outblkbuf[i] := bitdata[40+read_bits_asm(bits)];
      }
      if bits = 1 then
        decode_bitstream_asm_1
      else if bits = 2 then
        decode_bitstream_asm_2
      else if bits = 3 then
        {halt(0)} decode_bitstream_asm_3
      else
        decode_bitstream_asm(bits);
    end;




    if blocktype = 6 then { einfach kopieren }
    begin

      if (blockheader and 128) <> 0 then block_transparent := true;

      move(blockdata^[blkpos], outblkbuf^[0], 64);
      inc(blkpos, 64);

    end;


    if blocktype = 14 then { rle }
    begin

      blockrepeat := (blockheader shr 4) and 7;
      if blockheader and 128 <> 0 then
      begin
        inc(blockrepeat, blockdata^[blkpos] shl 3);
        inc(blkpos);
      end;
      { blockrepeat steht auf null wenn der block 2x gesetzt werden soll }
      { daher inkrementieren }
      inc(blockrepeat);

      goto start; { nchsten block einlesen }

    end;


    if blocktype = 15 then { redundante bloecke }
    begin

      p5 := redblockdata_pointer^[slot * 4 + (blockheader shr 5) and 3];

      red_offset := blockdata^[blkpos];
      inc(blkpos);
      if blockheader and 128 <> 0 then
      begin
        inc(red_offset, 256 * blockdata^[blkpos]);
        inc(blkpos);
      end;

      (*** Hier nun wieder Blocktypen 0, 1 und 6, diesmal aus red_data ***)

      blockheader := redblockdata^[red_offset];
      inc(red_offset);
      blocktype := blockheader and 15;

      block_transparent := false;

      if blocktype = 0 then { einfarbig }
      begin

        if blockheader and 64 <> 0 then
        begin
          fillchar(outblkbuf^[0], 64, redblockdata^[red_offset]);
          {inc(red_offset);}
        end;

        if blockheader and 32 <> 0 then
          fillchar(outblkbuf^[0], 64, ((blockheader and 16) shr 4) * 255);

        if blockheader = 0 then block_transparent := true;

      end;


      if (blocktype = 1) or (blocktype = 2) then { bitweise 1, 2, 3, 4, 5 }
      begin

        if blocktype = 2 then block_transparent := true;

        bits := (blockheader shr 4) and 7;
        bitdatasize := bits shl 3;

        move(redblockdata^[red_offset], bitdata^[0], bitdatasize);
        inc(red_offset, bitdatasize);

        pal_offset := redblockdata^[red_offset];
        inc(red_offset);
        if blockheader and 128 <> 0 then
        begin
          inc(pal_offset, 256 * redblockdata^[red_offset]);
          {inc(red_offset);}
        end;


        move(paldata^[pal_offset], bitdata^[40]{pal[0]}, 1 shl bits);

        {bitpos := 0;}
        {
        for i := 0 to 63 do
          outblkbuf[i] := bitdata[40+read_bits_asm(bits)];
        }
        if bits = 1 then
          decode_bitstream_asm_1
        else if bits = 2 then
          decode_bitstream_asm_2
        else if bits = 3 then
          {halt(0)} decode_bitstream_asm_3
        else
          decode_bitstream_asm(bits);
      end;




      if blocktype = 6 then { einfach kopieren }
      begin

        if (blockheader and 128) <> 0 then block_transparent := true;
        move(redblockdata^[red_offset], outblkbuf^[0], 64);
        {inc(red_offset, 64);}

      end;


      (*******************************************************************)

    end;


    for i := 0 to blockrepeat do
    begin
      if block_transparent then
      begin
        if blockheader <> 0 then
          copy_block_transp_asm(blkcount mod 40, blkcount div 40);
      end
      else
        copy_block_asm(blkcount mod 40, blkcount div 40);

      inc(blkcount);
    end;

  until blkcount = frame_blocks;

end;


end.